
DEPTH .

__( Cross compiling kernel )
ONLY CROSS ALSO DEFINITIONS

HEX

0000 T,  ( Sysinit jump loc )
0000 T,  ( Segment loader leave addr )
0000 T,  ( Offset loader leave addr )
0000 T,  ( Default DSP )
0000 T,  ( Default RSP )

VARIABLE BOOT
VARIABLE STATE

VARIABLE HEREP        ( HIDDEN )
VARIABLE LASTINSTR    ( 0000 LASTINSTR ! )

VARIABLE TIBLOC
VARIABLE TIB#
VARIABLE >IN
VARIABLE CHARBUF
VARIABLE SOURCE       ( 0000 SOURCE ! )

VARIABLE BASE
CREATE INCLSTACK      10 ALLOT
( VARIABLE INCLSTACK 0 , 0 , 0 , 0
VARIABLE INCLSP       T[ INCLSTACK INCLSP T! T]

THERE 0000 T,         ( Root vocabulary )
VARIABLE ROOTVOC      T[ TDUP ROOTVOC T! T]
VARIABLE USEVOC       T[ TDUP USEVOC T! T]
VARIABLE CURVOC       ( 0000 CURVOC ! )
VARIABLE DEFVOC       T[ DEFVOC T! T]
VARIABLE FIXUP        ( 0000 FIXUP ! )

0000 CONSTANT FLAG_PROCEDURE
0001 CONSTANT FLAG_IMMEDIATE
0002 CONSTANT FLAG_VARIABLE
0003 CONSTANT FLAG_CONSTANT
0080 CONSTANT FLAG_HIDDEN

0000 CONSTANT FALSE
FFFF CONSTANT TRUE


__( Building memory control )
TCELL CONSTANT CELL
: HERE  HEREP @ ;
: CELL+ CELL + ;
: CELLS CELL * ;
: ALLOT HEREP +! ;
:  , HERE  !    CELL ALLOT ;
: C, HERE C! 1 T_LIT ALLOT ;

__( Building heap control )
: I,     DUP LASTINSTR !  , ;
: IC,    DUP LASTINSTR ! C, ;
: CLROPT 0000 T_LIT LASTINSTR ! ;
: WASINSTR ( instr -- fl  ; Test if instr was last instruction )
   LASTINSTR @ = STATE @ AND ;

__( Building flags )
: OFF  0 T_LIT SWAP ! ;
: ON  -1 T_LIT SWAP ! ;

ONLY CROSS ALSO TARGET DEFINITIONS

__( Building vocabularies )
: ONLY         0000 T_LIT USEVOC ! ;
: FORTH        ROOTVOC @ CURVOC ! ;
: ALSO         CURVOC @ USEVOC ! ;
: DEFINITIONS  CURVOC @ DEFVOC ! ;

ONLY FORTH ALSO DEFINITIONS
CPUFILENAME @ COUNT FILEINCL
ONLY CROSS ALSO DEFINITIONS

__( Building interpreter support )

: ]
   STATE ON ; IMMEDIATE
: [
   STATE OFF ; IMMEDIATE

ONLY FORTH ALSO DEFINITIONS
IOFILENAME @ COUNT FILEINCL
ONLY CROSS ALSO DEFINITIONS

__( Building BEEP )
: BEEP ( --  ; Beep )
   0007 T_LIT EMIT ;

__( Building CR )
: CR   ( --  ; Next line )
   000D T_LIT EMIT 000A T_LIT EMIT ;

: COUNT ( str -- addr len  ; convert counted string to addr+len)
   DUP 1+ SWAP C@ ;

: TYPE ( addr len --  ; Output buffer )
   1 T_LIT WRITE-FILE ;

: TYPES ( str --  ; Output string )
   COUNT TYPE ;

: SAVE ( name len -- ; Save system to disk )
   0 T_LIT CREATE-FILE IF
      C" Error creating image" TYPES CR DROP
   ELSE
      DUP 0 T_LIT HERE ROT WRITE-FILE CLOSE-FILE DROP
      C" Image written" TYPES CR
   ENDIF ;

: PUSHINCL ( filehndl -- )
   INCLSP @ ! INCLSP @ CELL + INCLSP ! ;

: POPINCL ( -- filehndl )
   INCLSP @ CELL - DUP INCLSP ! @ ;

: CHARINPUT ( -- n1 ; Read char from default input found in SOURCE )
   SOURCE @ 0<> IF
      CHARBUF 1 T_LIT SOURCE @ READ-FILE DROP 0= IF ( EOF? )
         SOURCE @ CLOSE-FILE DROP POPINCL SOURCE ! 000D T_LIT
      ELSE
         CHARBUF @
      ENDIF
   ELSE
      KEY
   ENDIF ;

__( Building EXPECT )
: EXPECT ( buf maxlen -- len  ; Read line from console to buf )
   0000 T_LIT ROT ( -- maxlen len buf )
   BEGIN CHARINPUT ( -- maxlen len buf key )
      DUP 000D T_LIT <> WHILE
      DUP 000A T_LIT <> IF ( Skip LineFeeds in input )
         DUP 0008 T_LIT = IF
            2SWAP DUP 0= IF
               2SWAP BEEP DROP
            ELSE
               1- 2SWAP DUP EMIT 0020 T_LIT EMIT EMIT
            ENDIF
         ELSE
            2SWAP 2DUP = IF ( len=maxlen? )
               2SWAP BEEP DROP
            ELSE
               2SWAP SOURCE @ 0= IF DUP EMIT ENDIF ( -- maxlen len buf key )
               >R 2DUP + R> SWAP C! ( -- maxlen len buf )
               SWAP 1+ SWAP  ( -- maxlen len+1 buf )
            ENDIF
         ENDIF
      ELSE DROP
      ENDIF
   REPEAT ( -- maxlen len buf key )
   DROP DROP NIP SOURCE @ 0= IF 0020 T_LIT EMIT ENDIF ;

__( Building QUERY )
: QUERY ( --  ; Get inputline and write it to TIB )
   FF80 T_LIT TIBLOC ! TIBLOC @ 004E T_LIT
   EXPECT TIB# ! 0 T_LIT >IN ! ;

__( Building TOUPPER )
: TOUPPER ( ch -- ch  ; Convert char to uppercase )
   DUP ASCII a >= OVER ASCII z <= AND IF 0020 T_LIT - ENDIF ;
__( Building TOLOWER )
: TOLOWER ( ch -- ch  ; Convert char to lowercase )
   DUP ASCII A >= OVER ASCII Z <= AND IF 0020 T_LIT + ENDIF ;

__( Building 2INC )
: 2INC ( n1 n2 -- n1+1 n2+1  ; update 2 pointers )
   SWAP 1+ SWAP 1+ ;

__( Building MOVE )
: MOVE ( addr1 addr2 len --  ; move 'len' bytes from addr1 to addr2 )
   BEGIN DUP 0<> WHILE     ( -- src dest len )
   1- -ROT OVER C@ OVER C! ( -- len src dest  ; copy 1 char )
   2INC ROT REPEAT 3DROP ;

__( Building BUF2STR )
: BUF2STR ( src dest len -- ; move 'len' bytes from src to cstr dest )
   2DUP SWAP C!        ( -- src dest len  ; write string len )
   2DUP + 1+ 0 T_LIT SWAP C! ( -- src dest len ; write 0 term )
   -ROT 1+ ROT         ( -- src dest len  ; skip countbyte when copying )
   MOVE ;

__( Building WORDTIBPOS )
: WORDTIBPOS ( -- curtibpos  ; get current loc in TIB )
   TIBLOC @ >IN @ + ; HIDDEN

__( Building WORDSKIP )
: WORDSKIP ( char -- char  ; skip all char's in TIB )
   BEGIN
      >IN @ TIB# @ < WHILE
      WORDTIBPOS C@ OVER = WHILE
      1 T_LIT >IN +!
   REPEAT ; HIDDEN

__( Building WORDLEN )
: WORDLEN ( sep -- sep len  ; Find length of 'word' )
   >IN @ SWAP BEGIN  ( pos sep -- )
   OVER TIB# @ < WHILE
   OVER TIBLOC @ + C@ OVER <> WHILE
   SWAP 1+ SWAP
   REPEAT SWAP >IN @ - ; HIDDEN

__( Building WORD)
: WORD ( sep -- here  ; Copy string from input to HERE skip sep )
   WORDSKIP WORDLEN NIP DUP ( -- len len )
   WORDTIBPOS ( -- len len src )
   HERE ROT  ( -- len src dest len )
   BUF2STR ( -- len  ; copy string )
   1+ >IN +! ( --  ; update >IN )
   HERE ; ( -- here  ; return location of string )

__( Building FINDCMP )
: FINDCMP ( name defname -- fl  ; Is name same as defname ? )
   1 T_LIT -ROT ( fl name defname )
   2DUP C@ SWAP C@ = IF
      DUP C@ ( fl name defname len )
      BEGIN -ROT ( fl len name defname )
         2INC 2DUP C@ TOUPPER SWAP C@ TOUPPER <> IF
            >R >R >R CLR R> R> R>
         ENDIF ROT 1- DUP 0=
      UNTIL 3DROP
   ELSE 2DROP CLR
   ENDIF ; HIDDEN

__( Building (FIND )
: (FIND ( addr name def -- addr name )
   0 T_LIT BEGIN
      SWAP @ SWAP 0= OVER 0<> AND WHILE
      2DUP 3 T_LIT + FINDCMP DUP ( -- addr name def fl fl )
      IF >R ROT DROP DUP -ROT R> ENDIF ( -- addr name def fl )
   REPEAT DROP
   ; HIDDEN

__( Building FIND )
: FIND ( name -- addr  ; Find procedure by name )
   0 T_LIT SWAP ( -- addr name )
   USEVOC @ DUP 0<> IF
      (FIND
   ELSE
      DROP
   ENDIF

   OVER 0= IF
      ROOTVOC @ (FIND
   ENDIF
   DROP ;

__( Building compiling primitives )
: HEADER ( type --  ; Make word header )
   HERE DEFVOC @ DUP @ , ! C, ( --  ; fix link, register type )
   0020 T_LIT WORD C@ 1+ ALLOT ; ( --  ; Register name )
: CALCCODEADDR ( defaddr -- codeaddr  ; calculate code start point )
   2+ 1+ DUP C@ + 1+ ; HIDDEN
: VARIABLE ( --  ; Create variable )
   FLAG_VARIABLE HEADER 0 T_LIT DUP , , ;
: CONSTANT ( n1 --  ; Create constant )
   FLAG_CONSTANT HEADER 0 T_LIT , , ;
: CREATE ( --  ; Create null storage variable )
   FLAG_VARIABLE HEADER 0 T_LIT , ;
: DODOES
   DEFVOC @ @ CALCCODEADDR ! ;
: DOES>
   STATE @ IF
      0 T_LIT LIT HERE T[ DODOES T_LIT T] COMPILE, R<>D RET
      CELL - HERE SWAP ! R<>D
   ELSE
      C" Error in DOES> not compiling." TYPES CR
   ENDIF ; IMMEDIATE

__( Building PAD )
: PAD ( -- addr  ; get adres of temp loc on the heap )
   HERE 0100 T_LIT + ;

__( Building ATOI# )
: ATOI# ( char -- n fl  ; Convert one char to number )
   TOUPPER DUP ASCII A >= IF 0007 T_LIT - ENDIF ASCII 0 -

   DUP 0 T_LIT >= OVER BASE @ < AND IF ( number ? )
      TRUE
   ELSE
      CLR DUP
   ENDIF ;

__( Building U_ATOI )
: U_ATOI ( addr len -- un fl  ; Convert to unsigned num, fl=true conv ok )
   >R >R 0 T_LIT TRUE R> R> ( -- num fl addr len )
   BEGIN
      DUP 0<>
   WHILE
     SWAP ( -- num fl len str+1 )
     DUP C@ ATOI# ( -- num fl len str n fl )
     IF
        >R 2SWAP SWAP BASE @ * R> + SWAP 2SWAP ( increase num )
     ELSE
        DROP 2SWAP CLR 2SWAP ( clear ok flag )
     ENDIF 1+ SWAP 1-
   REPEAT 2DROP ;

__( Building ATOI )
: ATOI ( str -- n fl  ; Convert string to number, fl=true conv ok )
   COUNT OVER C@ ASCII - = IF  ( negative? )
      SWAP 1+ SWAP 1- U_ATOI SWAP NEGATE SWAP
   ELSE
      U_ATOI
   ENDIF ;

__( Building RUNWORD )
: RUNWORD ( defaddr --  ; Compile assembly code for current word )
   DUP 2+ C@ 007F T_LIT AND ( -- defaddr type&127 )

   DUP FLAG_PROCEDURE = IF
      DROP CALCCODEADDR @
      STATE @ IF COMPILE, ELSE EXECUTE ENDIF
   ELSE DUP FLAG_IMMEDIATE = IF
      DROP CALCCODEADDR @ EXECUTE
   ELSE DUP FLAG_VARIABLE = IF
      DROP CALCCODEADDR DUP 2+
      STATE @ IF
          LIT @ DUP 0<> IF COMPILE, ELSE DROP ENDIF
      ELSE
          SWAP @ DUP 0<> IF EXECUTE ELSE DROP ENDIF
      ENDIF
   ELSE DUP FLAG_CONSTANT = IF
      DROP CALCCODEADDR 2+ @
      STATE @ IF LIT ENDIF
   ELSE
      C" Unknown word type" TYPES CR
      DROP
   ENDIF ENDIF ENDIF ENDIF ;

__( Building RUN )
: RUN ( --  ; Interpret input and process )
   BEGIN >IN @ TIB# @ < WHILE
      0020 T_LIT WORD DUP DUP C@ 0<> IF
         FIND DUP 0= IF
( C"  Not found" TYPES CR
            ( word not found, try as literal )
            DROP DUP ATOI IF
               NIP STATE @ IF LIT ENDIF ( compile/push number )
            ELSE
               DROP CR TYPES C"  <- Wrong keyword " TYPES
            ENDIF
         ELSE
( C"  Found" TYPES CR
            ( word found, compile and/or execute )
            NIP RUNWORD
         ENDIF
      ENDIF
   REPEAT ;

: PROMPT
   C"  ok>" TYPES CR ;

: ?PROMPT
   SOURCE @ 0= IF STATE @ IF CR ELSE PROMPT ENDIF ENDIF ;

: QUIT ( The QUIT interpreter loop )
   STATE OFF
   BEGIN
      QUERY RUN ?PROMPT
   REPEAT ;

: FILEINCL ( name len -- )
   SOURCE @ PUSHINCL
   0 T_LIT OPEN-FILE IF
      C" Error opening file" TYPES CR DROP
   ELSE
      SOURCE ! ( QUIT )
   ENDIF ;

: INCLUDE ( |name -- )
   0020 T_LIT WORD COUNT FILEINCL ;

: SAVEAS
   0020 T_LIT WORD COUNT SAVE ;

: BOOTMSG
   CR C" MPForth 0.9.729 PW.Soft. 1999" TYPES CR ;

__( Building BOOT )
: (BOOT ( --  ; boot routine )
   BOOTMSG           ( Display Boot message )
   BEGIN QUIT REPEAT ( Start interpreter )
   ;

: HEX     0010 T_LIT BASE ! ;
: DECIMAL 000A T_LIT BASE ! ;
: OCTAL   0008 T_LIT BASE ! ;
: BIN     0002 T_LIT BASE ! ;


ONLY CROSS ALSO TARGET DEFINITIONS


: .( ASCII ) WORD TYPES ; IMMEDIATE
: ( ASCII ) WORD DROP ; IMMEDIATE
: C"
   JMPFWD
   ASCII " WORD ( -- afix here )
   DUP C@ 1+ ALLOT
   0 T_LIT C, ( -- afix here  ; null terminate string )
   SWAP FIXFWD RELOC LIT ; IMMEDIATE
: ' ( | name -- cfa  ; Search word and provide cfa )
   0020 T_LIT WORD FIND CALCCODEADDR @ ;
: ['] ( | name -- cfa ; Search word and provide cfa )
   0020 T_LIT WORD FIND CALCCODEADDR @ LIT ; IMMEDIATE
: ASCII ( -- ch )
   0020 T_LIT WORD 1+ C@
   STATE @ IF LIT ENDIF ; IMMEDIATE

: IMMEDIATE ( --  ; Change wordtype to imm )
   DEFVOC @ @ 2+ DUP C@ FLAG_IMMEDIATE OR SWAP C! ;
: HIDDEN ( --  ; Change wordtype to imm )
   DEFVOC @ @ 2+ DUP C@ FLAG_HIDDEN OR SWAP C! ;
: :
   FLAG_PROCEDURE HEADER HERE CELL+ , R<>D ] ;
: ;
   R<>D RET [ ; IMMEDIATE


ONLY FORTH ALSO DEFINITIONS
SYSINITFILENAME @ COUNT FILEINCL
ONLY CROSS ALSO DEFINITIONS


( Setup default configuration parameters )
T[
   000A BASE T!
   (BOOT BOOT T!
   TVOC ROOTVOC T@ T!
T]
DEPTH .

