
#include <stdlib.h>
#include <fcntl.h>
#include <dos.h>
#include <io.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>

#define MAXIMAGESIZE     64000
#define DATASTACKSIZE     8192
#define RETURNSTACKSIZE   8192

typedef unsigned char _U8;
typedef unsigned long _U32;


_U8  *ImageMem=NULL;
_U32 *DataStack=NULL;
_U32 *ReturnStack=NULL;

_U32 _TOS;
_U32 _DSP;
_U32 _RSP;
_U32 _IP;
int RunningFlag;



//
// Reserve memory
//
int ReserveMem(void) {
	int Retval=-1;

	ImageMem   =(_U8 *)malloc(MAXIMAGESIZE);
	DataStack  =(_U32 *)malloc(DATASTACKSIZE);
	ReturnStack=(_U32 *)malloc(RETURNSTACKSIZE);

	if (ImageMem && DataStack && ReturnStack) {
		Retval=0;
	}

	return Retval;
}



//
// Free reserved memory
//
int FreeMem(void) {
	if (ImageMem) {
		free(ImageMem);
		ImageMem=NULL;
	}
	if (DataStack) {
		free(DataStack);
		DataStack=NULL;
	}
	if (ReturnStack) {
		free(ReturnStack);
		ReturnStack=NULL;
	}
	return 0;
}



//
// Display error to screen
//
int Error(const char *errstr) {
	int i;

	_IP--;
	fprintf(stderr,"\n\nERROR: %s\n"
		"_IP=0x%08lX TOS=0x%08lX DSP=0x%08lX RSP=0x%08lX\nCODE:"
		,errstr,_IP,_TOS,_DSP,_RSP);
	for(i=0;i<16;i++) {
		printf(" %02X",ImageMem[_IP+i]);
	}
	printf("\nDSTK:");
	for(i=0;i<8;i++) {
		if (_DSP+i>=8) printf(" %08lX",DataStack[(_DSP-8)+i]);
	}
	printf("\n");
	RunningFlag=0;
	return 0;
}


//
//
//
inline _U32 Get32(_U32 addr) {
	return *((_U32 *)(ImageMem+addr));
}

//
//
//
inline void Put32(_U32 addr,_U32 value) {
	printf("Put debug: %08lx <- %08lx\n",addr,value);
	*((_U32 *)(ImageMem+addr))=value;
}


//
//
//
inline void WRITE_FILE(void) {
	printf("debug write file(%08lx,%08lx,%08lx)\n",DataStack[_DSP-2],DataStack[_DSP-1],_TOS);
	write(_TOS,ImageMem+DataStack[_DSP-2],DataStack[_DSP-1]);
	_DSP-=3;
	_TOS=DataStack[_DSP];
}


//
// Boot loaded image
//
int BootImage(void) {
	_U32 tmp;

	_TOS=0;
	_DSP=0;
	_RSP=0;
	_IP=Get32(0);
	RunningFlag=1;

	while(RunningFlag) {
//		printf("TRACE IP=%08lx INSTR=%02X",_IP,ImageMem[_IP]);
		switch(ImageMem[_IP++]) {
		case 0x00:
			printf("Instr 0\n");
			break;
		case 0x01:
			printf("Instr 1\n");
			break;
		case 0x10: /* DUP */
			DataStack[_DSP++]=_TOS;
			break;
		case 0x11: /* DROP */
			_TOS=DataStack[--_DSP];
			break;
		case 0x12: /* SWAP */
			tmp=_TOS;
			_TOS=DataStack[_DSP-1];
			DataStack[_DSP-1]=tmp;
			break;
		case 0x13: /* NIP */
			break;
		case 0x14: /* OVER */
			break;
		case 0x20: /* LIT */
			DataStack[_DSP++]=_TOS;
			_TOS=Get32(_IP);
			_IP+=4;
			break;
		case 0x21:
			_TOS=Get32(_TOS);
			break;
		case 0x22:
			printf("PUT: ");
			Put32(_TOS,DataStack[--_DSP]);
			_TOS=DataStack[--_DSP];
			break;
		case 0x23: /* GET8 */
			_TOS=ImageMem[_TOS];
			break;
		case 0x30: /* 1+ */
			_TOS++;
			break;
		case 0x70:
			DataStack[_DSP++]=_TOS;
			_TOS=_DSP;
			break;	
		case 0x71:
			DataStack[_DSP++]=_TOS;
			_TOS=_RSP;
			break;
		case 0x78: /* JMPTOS */
			_IP=_TOS;
			_TOS=DataStack[--_DSP];
			break;
		case 0x79: /* JMPIMM */
			_IP=Get32(_IP);
			break;
		case 0x7A: /* JSRTOS */
			ReturnStack[_RSP++]=_IP+4;
			_IP=_TOS;
			_TOS=DataStack[--_DSP];
			break;
		case 0x7B: /* JSRIMM */
			ReturnStack[_RSP++]=_IP+4;
			_IP=Get32(_IP);
			break;
		case 0x7F: /* RETURN */
			_IP=ReturnStack[--_RSP];
			break;
		case 0x82: /* WRITE-FILE addr count hndl */
			WRITE_FILE();
			break;
		default:
			Error("Invalid instruction");
			break;
		}
	}

	return 0;
}



int LoadImage(const char *imagefilename) {
	int Retval=-1;
	int fhandle=open(imagefilename,O_RDONLY);

	if (fhandle>=0) {
		Retval=read(fhandle,ImageMem,MAXIMAGESIZE);
	}

	return Retval;
}



int main(int argc,char **argv) {
	int Retval=10;
	int imagesize;

	if (argc<2) {
		printf("4THLOAD 0.9.720 - MPForth interpreter\n"
                       "Usage: 4THLOAD <image>\n");
	} else {
		if (ReserveMem()==0) {
			imagesize=LoadImage(argv[1]);
			if (imagesize>0) {
				printf("Image loaded\n");
				BootImage();
			} else {
				printf("Error loading image\n");
			}
		} else {
			printf("Error reserving memory\n");
		}
		FreeMem();
	}

	return Retval;
}

